@@ -1,19 +1,18 @@
 import warnings
-from collections import OrderedDict
 from collections.abc import Mapping, Sequence
-from functools import partial
 from ipaddress import ip_address
 from urllib.parse import (SplitResult, parse_qsl,
                           urljoin, urlsplit, urlunsplit)
 
 from multidict import MultiDict, MultiDictProxy
+import idna
 
 
-from .quoting import quote, unquote
+from .quoting import _Quoter, _Unquoter
 
-__version__ = '0.14.2'
+__version__ = '1.2.2'
 
-__all__ = ['URL']
+__all__ = ('URL',)
 
 
 # is_leaf()
@@ -26,13 +25,8 @@ DEFAULT_PORTS = {
     'wss': 443,
 }
 
-PROTECT_CHARS = '=+&;'
-
 sentinel = object()
 
-_quote = quote
-_unquote = unquote
-
 
 class cached_property:
     """Use as a class method decorator.  It operates almost exactly like
@@ -137,16 +131,24 @@ class URL:
     # absolute-URI  = scheme ":" hier-part [ "?" query ]
     __slots__ = ('_cache', '_val')
 
-    def __init__(self, val='', *, encoded=False, strict=None):
+    _QUOTER = _Quoter()
+    _PATH_QUOTER = _Quoter(safe='@:', protected='/+')
+    _QUERY_QUOTER = _Quoter(safe='?/:@', protected='=+&;', qs=True)
+    _QUERY_PART_QUOTER = _Quoter(safe='?/:@', qs=True)
+    _FRAGMENT_QUOTER = _Quoter(safe='?/:@')
+
+    _UNQUOTER = _Unquoter()
+    _PATH_UNQUOTER = _Unquoter(unsafe='+')
+    _QS_UNQUOTER = _Unquoter(qs=True)
+
+    def __new__(cls, val='', *, encoded=False, strict=None):
         if strict is not None:  # pragma: no cover
             warnings.warn("strict parameter is ignored")
-        if isinstance(val, URL):
-            self._val = val._val
-            self._cache = val._cache
-            return
-        if isinstance(val, str):
+        if type(val) is cls:
+            return val
+        if type(val) is str:
             val = urlsplit(val)
-        elif isinstance(val, SplitResult):
+        elif type(val) is SplitResult:
             if not encoded:
                 raise ValueError("Cannot apply decoding to SplitResult")
         else:
@@ -156,48 +158,32 @@ class URL:
             if not val[1]:  # netloc
                 netloc = ''
             else:
-                netloc = val.hostname
-                if netloc is None:
+                host = val.hostname
+                if host is None:
                     raise ValueError(
                         "Invalid URL: host is required for abolute urls.")
-                try:
-                    netloc.encode('ascii')
-                except UnicodeEncodeError:
-                    netloc = netloc.encode('idna').decode('ascii')
-                else:
-                    try:
-                        ip = ip_address(netloc)
-                    except ValueError:
-                        pass
-                    else:
-                        if ip.version == 6:
-                            netloc = '[' + netloc + ']'
-                if val.port:
-                    netloc += ':{}'.format(val.port)
-                if val.username:
-                    user = _quote(val.username)
-                else:
-                    user = ''
-                if val.password:
-                    user += ':' + _quote(val.password)
-                if user:
-                    netloc = user + '@' + netloc
-
-            path = _quote(val[2], safe='+@:', protected='/+')
+                netloc = cls._make_netloc(val.username,
+                                          val.password,
+                                          host,
+                                          val.port,
+                                          encode=True)
+            path = cls._PATH_QUOTER(val[2])
             if netloc:
-                path = _normalize_path(path)
+                path = cls._normalize_path(path)
 
-            query = _quote(val[3], safe='=+&?/:@',
-                           protected=PROTECT_CHARS, qs=True)
-            fragment = _quote(val[4], safe='?/:@')
+            query = cls._QUERY_QUOTER(val[3])
+            fragment = cls._FRAGMENT_QUOTER(val[4])
             val = SplitResult(val[0], netloc, path, query, fragment)
 
+        self = object.__new__(cls)
         self._val = val
         self._cache = {}
+        return self
 
     @classmethod
     def build(cls, *, scheme='', user='', password='', host='', port=None,
-              path='', query=None, query_string='', fragment=''):
+              path='', query=None, query_string='', fragment='',
+              encoded=False):
         """Creates and returns a new URL"""
 
         if host and not scheme:
@@ -210,17 +196,25 @@ class URL:
             raise ValueError(
                 "Only one of \"query\" or \"query_string\" should be passed")
 
-        netloc = cls._make_netloc(user, password, host, port)
-        path = _quote(path, safe='@:', protected='/')
-        if netloc:
-            path = _normalize_path(path)
+        if not user and not password and not host and not port:
+            netloc = ''
+        else:
+            netloc = cls._make_netloc(user, password, host, port,
+                                      encode=not encoded)
+        if not encoded:
+            path = cls._PATH_QUOTER(path)
+            if netloc:
+                path = cls._normalize_path(path)
+
+            query_string = cls._QUERY_QUOTER(query_string)
+            fragment = cls._FRAGMENT_QUOTER(fragment)
 
         url = cls(
             SplitResult(
                 scheme,
                 netloc,
                 path,
-                _quote(query_string),
+                query_string,
                 fragment
             ),
             encoded=True
@@ -231,6 +225,10 @@ class URL:
         else:
             return url
 
+    def __init_subclass__(cls):
+        raise TypeError("Inheritance a class {!r} from URL "
+                        "is forbidden".format(cls))
+
     def __str__(self):
         val = self._val
         if not val.path and self.is_absolute() and (val.query or val.fragment):
@@ -241,7 +239,7 @@ class URL:
         return "{}('{}')".format(self.__class__.__name__, str(self))
 
     def __eq__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
 
         val1 = self._val
@@ -264,27 +262,27 @@ class URL:
         return ret
 
     def __le__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val <= other._val
 
     def __lt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val < other._val
 
     def __ge__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val >= other._val
 
     def __gt__(self, other):
-        if not isinstance(other, URL):
+        if not type(other) is URL:
             return NotImplemented
         return self._val > other._val
 
     def __truediv__(self, name):
-        name = _quote(name, safe=':@', protected='/')
+        name = self._PATH_QUOTER(name)
         if name.startswith('/'):
             raise ValueError("Appending path "
                              "starting from slash is forbidden")
@@ -298,7 +296,7 @@ class URL:
             parts.append(name)
             new_path = '/'.join(parts)
         if self.is_absolute():
-            new_path = _normalize_path(new_path)
+            new_path = self._normalize_path(new_path)
         return URL(self._val._replace(path=new_path, query='', fragment=''),
                    encoded=True)
 
@@ -349,7 +347,8 @@ class URL:
         if not self._val.scheme:
             raise ValueError("URL should have scheme")
         v = self._val
-        netloc = self._make_netloc(None, None, v.hostname, v.port)
+        netloc = self._make_netloc(None, None, v.hostname, v.port,
+                                   encode=False)
         val = v._replace(netloc=netloc, path='', query='', fragment='')
         return URL(val, encoded=True)
 
@@ -393,7 +392,7 @@ class URL:
         None if user is missing.
 
         """
-        return _unquote(self.raw_user)
+        return self._UNQUOTER(self.raw_user)
 
     @property
     def raw_password(self):
@@ -411,7 +410,7 @@ class URL:
         None if password is missing.
 
         """
-        return _unquote(self.raw_password)
+        return self._UNQUOTER(self.raw_password)
 
     @property
     def raw_host(self):
@@ -434,7 +433,16 @@ class URL:
         raw = self.raw_host
         if raw is None:
             return None
-        return raw.encode('ascii').decode('idna')
+        if '%' in raw:
+            # Hack for scoped IPv6 addresses like
+            # fe80::2%Проверка
+            # presence of '%' sign means only IPv6 address, so idna is useless.
+            return raw
+
+        try:
+            return idna.decode(raw.encode('ascii'))
+        except UnicodeError:  # e.g. '::1'
+            return raw.encode('ascii').decode('idna')
 
     @property
     def port(self):
@@ -465,7 +473,7 @@ class URL:
         / for absolute URLs without path part.
 
         """
-        return _unquote(self.raw_path, unsafe='+')
+        return self._PATH_UNQUOTER(self.raw_path)
 
     @cached_property
     def query(self):
@@ -495,16 +503,22 @@ class URL:
         Empty string if query is missing.
 
         """
-        return _unquote(self.raw_query_string, qs=True)
+        return self._QS_UNQUOTER(self.raw_query_string)
 
     @cached_property
     def path_qs(self):
-        """Decoded path of URL with query
-        """
+        """Decoded path of URL with query."""
         if not self.query_string:
             return self.path
         return '{}?{}'.format(self.path, self.query_string)
 
+    @cached_property
+    def raw_path_qs(self):
+        """Encoded path of URL with query."""
+        if not self.raw_query_string:
+            return self.raw_path
+        return '{}?{}'.format(self.raw_path, self.raw_query_string)
+
     @property
     def raw_fragment(self):
         """Encoded fragment part of URL.
@@ -521,7 +535,7 @@ class URL:
         Empty string if fragment is missing.
 
         """
-        return _unquote(self.raw_fragment)
+        return self._UNQUOTER(self.raw_fragment)
 
     @cached_property
     def raw_parts(self):
@@ -550,7 +564,7 @@ class URL:
         ('/',) for absolute URLs if *path* is missing.
 
         """
-        return tuple(_unquote(part) for part in self.raw_parts)
+        return tuple(self._UNQUOTER(part) for part in self.raw_parts)
 
     @cached_property
     def parent(self):
@@ -585,17 +599,74 @@ class URL:
     @cached_property
     def name(self):
         """The last part of parts."""
-        return _unquote(self.raw_name)
+        return self._UNQUOTER(self.raw_name)
 
     @classmethod
-    def _make_netloc(cls, user, password, host, port):
-        ret = host
+    def _normalize_path(cls, path):
+        # Drop '.' and '..' from path
+
+        segments = path.split('/')
+        resolved_path = []
+
+        for seg in segments:
+            if seg == '..':
+                try:
+                    resolved_path.pop()
+                except IndexError:
+                    # ignore any .. segments that would otherwise cause an
+                    # IndexError when popped from resolved_path if
+                    # resolving for rfc3986
+                    pass
+            elif seg == '.':
+                continue
+            else:
+                resolved_path.append(seg)
+
+        if segments[-1] in ('.', '..'):
+            # do some post-processing here.
+            # if the last segment was a relative dir,
+            # then we need to append the trailing '/'
+            resolved_path.append('')
+
+        return '/'.join(resolved_path)
+
+    @classmethod
+    def _encode_host(cls, host):
+        try:
+            ip, sep, zone = host.partition('%')
+            ip = ip_address(ip)
+        except ValueError:
+            try:
+                host = idna.encode(host, uts46=True).decode('ascii')
+            except UnicodeError:
+                host = host.encode('idna').decode('ascii')
+        else:
+            host = ip.compressed
+            if sep:
+                host += '%' + zone
+            if ip.version == 6:
+                host = '[' + host + ']'
+        return host
+
+    @classmethod
+    def _make_netloc(cls, user, password, host, port, encode):
+        if encode:
+            ret = cls._encode_host(host)
+        else:
+            ret = host
         if port:
             ret = ret + ':' + str(port)
         if password:
             if not user:
                 user = ''
+            else:
+                if encode:
+                    user = cls._QUOTER(user)
+            if encode:
+                password = cls._QUOTER(password)
             user = user + ':' + password
+        elif user and encode:
+            user = cls._QUOTER(user)
         if user:
             ret = user + '@' + ret
         return ret
@@ -623,7 +694,7 @@ class URL:
         if user is None:
             password = None
         elif isinstance(user, str):
-            user = _quote(user)
+            user = self._QUOTER(user)
             password = val.password
         else:
             raise TypeError("Invalid user type")
@@ -633,7 +704,8 @@ class URL:
         return URL(self._val._replace(netloc=self._make_netloc(user,
                                                                password,
                                                                val.hostname,
-                                                               val.port)),
+                                                               val.port,
+                                                               encode=False)),
                    encoded=True)
 
     def with_password(self, password):
@@ -648,7 +720,7 @@ class URL:
         if password is None:
             pass
         elif isinstance(password, str):
-            password = _quote(password)
+            password = self._QUOTER(password)
         else:
             raise TypeError("Invalid password type")
         if not self.is_absolute():
@@ -660,7 +732,8 @@ class URL:
                 netloc=self._make_netloc(val.username,
                                          password,
                                          val.hostname,
-                                         val.port)),
+                                         val.port,
+                                         encode=False)),
             encoded=True)
 
     def with_host(self, host):
@@ -680,19 +753,14 @@ class URL:
                              "for relative URLs")
         if not host:
             raise ValueError("host removing is not allowed")
-        try:
-            ip = ip_address(host)
-        except ValueError:
-            host = host.encode('idna').decode('ascii')
-        else:
-            if ip.version == 6:
-                host = '[' + host + ']'
+        host = self._encode_host(host)
         val = self._val
         return URL(
             self._val._replace(netloc=self._make_netloc(val.username,
                                                         val.password,
                                                         host,
-                                                        val.port)),
+                                                        val.port,
+                                                        encode=False)),
             encoded=True)
 
     def with_port(self, port):
@@ -713,15 +781,16 @@ class URL:
             self._val._replace(netloc=self._make_netloc(val.username,
                                                         val.password,
                                                         val.hostname,
-                                                        port)),
+                                                        port,
+                                                        encode=False)),
             encoded=True)
 
     def with_path(self, path, *, encoded=False):
         """Return a new URL with path replaced."""
         if not encoded:
-            path = _quote(path, safe='@:', protected='/')
+            path = self._PATH_QUOTER(path)
             if self.is_absolute():
-                path = _normalize_path(path)
+                path = self._normalize_path(path)
         if len(path) > 0 and path[0] != '/':
             path = '/' + path
         return URL(self._val._replace(path=path, query='', fragment=''),
@@ -742,7 +811,7 @@ class URL:
         if query is None:
             query = ''
         elif isinstance(query, Mapping):
-            quoter = partial(_quote, qs=True)
+            quoter = self._QUERY_PART_QUOTER
             lst = []
             for k, v in query.items():
                 if isinstance(v, str):
@@ -753,17 +822,15 @@ class URL:
                     raise TypeError("Invalid variable type: mapping value "
                                     "should be str or int, got {!r}".format(v))
                 lst.append(
-                    quoter(k, safe='/?:@') + '=' + quoter(v, safe='/?:@'))
-                query = '&'.join(lst)
+                    quoter(k) + '=' + quoter(v))
+            query = '&'.join(lst)
         elif isinstance(query, str):
-            query = _quote(query, safe='/?:@',
-                           protected=PROTECT_CHARS,
-                           qs=True)
+            query = self._QUERY_QUOTER(query)
         elif isinstance(query, (bytes, bytearray, memoryview)):
             raise TypeError("Invalid query type: bytes, bytearray and "
                             "memoryview are forbidden")
         elif isinstance(query, Sequence):
-            quoter = partial(_quote, qs=True, safe='/?:@')
+            quoter = self._QUERY_PART_QUOTER
             query = '&'.join(quoter(k) + '=' + quoter(v)
                              for k, v in query)
         else:
@@ -792,19 +859,13 @@ class URL:
 
     def update_query(self, *args, **kwargs):
         """Return a new URL with query part updated."""
-        new_query = OrderedDict(
-            map(
-                lambda x: x.split('=', 1),
-                _quote(self._get_str_query(*args, **kwargs),
-                       safe='/?:@', protected=PROTECT_CHARS,
-                       qs=True).lstrip("?").split("&")
-            )
-        )
-        query = OrderedDict(self.query)
+        s = self._get_str_query(*args, **kwargs)
+        new_query = MultiDict(parse_qsl(s, keep_blank_values=True))
+        query = MultiDict(self.query)
         query.update(new_query)
-        query = self._get_str_query(query)
-        return URL(
-            self._val._replace(path=self._val.path, query=query), encoded=True)
+
+        return URL(self._val._replace(query=self._get_str_query(query)),
+                   encoded=True)
 
     def with_fragment(self, fragment):
         """Return a new URL with fragment replaced.
@@ -821,7 +882,7 @@ class URL:
             raise TypeError("Invalid fragment type")
         return URL(
             self._val._replace(
-                fragment=_quote(fragment, safe='?/:@')),
+                fragment=self._FRAGMENT_QUOTER(fragment)),
             encoded=True)
 
     def with_name(self, name):
@@ -837,7 +898,7 @@ class URL:
             raise TypeError("Invalid name type")
         if '/' in name:
             raise ValueError("Slash in name is not allowed")
-        name = _quote(name, safe='@:', protected='/')
+        name = self._PATH_QUOTER(name)
         if name in ('.', '..'):
             raise ValueError(". and .. values are forbidden")
         parts = list(self.raw_parts)
@@ -879,35 +940,8 @@ class URL:
                                       self._make_netloc(self.user,
                                                         self.password,
                                                         self.host,
-                                                        self._val.port),
+                                                        self._val.port,
+                                                        encode=False),
                                       self.path,
                                       self.query_string,
                                       self.fragment))
-
-
-def _normalize_path(path):
-    # Drop '.' and '..' from path
-
-    segments = path.split('/')
-    resolved_path = []
-
-    for seg in segments:
-        if seg == '..':
-            try:
-                resolved_path.pop()
-            except IndexError:
-                # ignore any .. segments that would otherwise cause an
-                # IndexError when popped from resolved_path if
-                # resolving for rfc3986
-                pass
-        elif seg == '.':
-            continue
-        else:
-            resolved_path.append(seg)
-
-    if segments[-1] in ('.', '..'):
-        # do some post-processing here. if the last segment was a relative dir,
-        # then we need to append the trailing '/'
-        resolved_path.append('')
-
-    return '/'.join(resolved_path)
